/**
 * \file input/xtandemsaxhandler.cpp
 * \date 5/4/2017
 * \author Olivier Langella
 * \brief parse XML X!Tandem result file
 */


/*******************************************************************************
* Copyright (c) 2017 Olivier Langella <Olivier.Langella@u-psud.fr>.
*
* This file is part of XTPcpp.
*
*     XTPcpp is free software: you can redistribute it and/or modify
*     it under the terms of the GNU General Public License as published by
*     the Free Software Foundation, either version 3 of the License, or
*     (at your option) any later version.
*
*     XTPcpp is distributed in the hope that it will be useful,
*     but WITHOUT ANY WARRANTY; without even the implied warranty of
*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*     GNU General Public License for more details.
*
*     You should have received a copy of the GNU General Public License
*     along with XTPcpp.  If not, see <http://www.gnu.org/licenses/>.
*
* Contributors:
*     Olivier Langella <Olivier.Langella@u-psud.fr> - initial API and implementation
******************************************************************************/

#include "xtandemsaxhandler.h"

#include <pappsomspp/exception/exceptionnotfound.h>
#include <cmath>
#include "../utils/peptidestore.h"
#include "../utils/proteinstore.h"

XtandemSaxHandler::XtandemSaxHandler(Project * p_project, IdentificationGroup * p_identification_group,
                                     IdentificationDataSource * p_identification_data_source):_p_project(p_project)
{
    _p_identification_group = p_identification_group;

    _p_identification_data_source =p_identification_data_source;
    _sp_msrun = p_identification_data_source->getMsRunSp();
}

XtandemSaxHandler::~XtandemSaxHandler()
{

}


bool XtandemSaxHandler::startElement(const QString & namespaceURI, const QString & localName,
                                     const QString & qName, const QXmlAttributes & attributes) {
    // qDebug() << namespaceURI << " " << localName << " " << qName ;
    _tag_stack.push_back(qName);
    bool is_ok = true;

    try {
        //startElement_group
        if (qName == "group") {
            is_ok = startElement_group(attributes);
        } else if (qName == "protein") {
            is_ok = startElement_protein(attributes);
        } else if (qName == "note") {
            is_ok = startElement_note(attributes);
        }
        else if (qName == "file") {
            is_ok = startElement_file(attributes);

        } else if (qName == "aa") {
            is_ok = startElement_aa(attributes);
        } else if (qName == "domain") {
            is_ok = startElement_domain(attributes);
        }

        _current_text.clear();
    }
    catch (pappso::PappsoException exception_pappso) {
        _errorStr = QObject::tr("ERROR in XtandemSaxHandler::startElement tag %1, PAPPSO exception:\n%2").arg(qName).arg(exception_pappso.qwhat());
        return false;
    }
    catch (std::exception exception_std) {
        _errorStr = QObject::tr("ERROR in XtandemSaxHandler::startElement tag %1, std exception:\n%2").arg(qName).arg(exception_std.what());
        return false;
    }
    return is_ok;
}

bool XtandemSaxHandler::endElement(const QString & namespaceURI, const QString & localName,
                                   const QString & qName) {

    bool is_ok = true;
    // endElement_peptide_list
    try {
        if (qName == "note")
        {
            is_ok = endElement_note();
        } else if (qName == "domain")
        {
            is_ok = endElement_domain();
        }

        // end of detection_moulon
        // else if ((_tag_stack.size() > 1) &&
        //         (_tag_stack[_tag_stack.size() - 2] == "detection_moulon"))
    }
    catch (pappso::PappsoException exception_pappso) {
        _errorStr = QObject::tr("ERROR in XtandemSaxHandler::endElement tag %1, PAPPSO exception:\n%2").arg(qName).arg(exception_pappso.qwhat());
        return false;
    }
    catch (std::exception exception_std) {
        _errorStr = QObject::tr("ERROR in XtandemSaxHandler::endElement tag %1, std exception:\n%2").arg(qName).arg(exception_std.what());
        return false;
    }

    _current_text.clear();
    _tag_stack.pop_back();

    return is_ok;
}


bool XtandemSaxHandler::startElement_group(QXmlAttributes attrs) {
    //<group id="1976" mh="1120.529471" z="2" rt="PT667.022S" expect="9.7e-04" label="GRMZM2G083841_P01 P04711 Phosphoenolpyruvate carboxylase 1 (PEPCase 1)(PEPC 1)(EC..." type="model" sumI="5.34" maxI="35986.9" fI="359.869" act="0" >
    bool is_ok = true;
    // logger.debug("startElementgroup begin");
    // <group label="performance parameters" type="parameters">
    _current_group_label = attrs.value("label");
    _current_group_type = attrs.value("type");
    if (_current_group_type == "model") {
        _scan = attrs.value("id").toUInt();
        _mhplus_obser = attrs.value("mh").toDouble();
        _charge = attrs.value("z").toUInt();
        //_retention_time = attrs.value("rt");
    }
    return is_ok;
}

bool XtandemSaxHandler::startElement_note(QXmlAttributes attributes) {
//<note label="description">GRMZM2G083841_P01 P04711 Phosphoenolpyruvate carboxylase 1 (PEPCase 1)(PEPC 1)(EC //4.1.1.31) seq=translation; coord=9:61296279..61301686:1; parent_transcript=GRMZM2G083841_T01;
    ////parent_gene=GRMZM2G083841</note>
    bool is_ok = true;
    _is_protein_description = false;
    if (attributes.value("label") == "description") {
        if (_tag_stack[_tag_stack.size() - 2] == "protein") {
            _is_protein_description = true;
        }
    }
    return is_ok;
}

bool XtandemSaxHandler::startElement_protein(QXmlAttributes attributes) {
//<protein expect="-704.6" id="1976.1" uid="195701" label="GRMZM2G083841_P01 P04711 Phosphoenolpyruvate carboxylase 1 (PEPCase 1)(PEPC 1)(EC..." sumI="9.36" >
    bool is_ok = true;
    //qDebug() << "startElement_protein begin";
    QString accession = attributes.value("label").simplified().split(" ", QString::SkipEmptyParts).at(0);
    //qDebug() << "startElement_protein accession" << accession;
    _p_protein_match = _p_identification_group->getProteinMatch(accession);

    //qDebug() << "startElement_protein p_protein_match 1 " << _p_protein_match;
    _current_protein.setAccession(accession);
    if (_p_protein_match == nullptr) {
        _p_protein_match = new ProteinMatch();
        //qDebug() << "startElement_protein p_protein_match 2 " << _p_protein_match;
        _p_protein_match->setChecked(false);

    }
    //qDebug() << "startElement_protein p_protein_match 3 " << _p_protein_match;
    ProteinXtpSp sp_xtp_protein = _current_protein.makeProteinXtpSp();
    _p_protein_match->setProteinXtpSp(_p_project->getProteinStore().getInstance(sp_xtp_protein));
    _p_protein_match->setChecked(true);

    _p_identification_group->addProteinMatch(_p_protein_match);
    //qDebug() << "startElement_protein end" ;
    return is_ok;
}


bool XtandemSaxHandler::startElement_file(QXmlAttributes attributes) {
    bool is_ok = true;
    //<file type="peptide" URL="/gorgone/pappso/formation/TD/Database/Genome_Z_mays_5a.fasta"/>
    if (attributes.value("type") == "peptide") {
        //prot_.setDatabase(identification_.getDatabaseSet().getInstance(
        //                     attrs.getValue("URL")));
    }
    return is_ok;
}

bool XtandemSaxHandler::startElement_domain(QXmlAttributes attributes) {
//<domain id="1976.1.1" start="620" end="629" expect="9.7e-04" mh="1120.5307" delta="-0.0012" hyperscore="29.9"
    //nextscore="10.2" y_score="10.4" y_ions="7" b_score="11.2" b_ions="3" pre="QLYR" post="RYGV"
    //seq="AQEEMAQVAK" missed_cleavages="0">
    //qDebug() << "startElement_domain ";
    bool is_ok = true;
    _current_text = _current_text.simplified().replace(" ", "");
    if (!_current_text.isEmpty()) {
        _p_protein_match->getProteinXtpSp().get()->setSequence(_current_text);
    }

    // <domain id="4017.1.1" start="21" end="31" expect="2.0e-06"
    // mh="1263.575"
    // delta="0.998" hyperscore="32.9" nextscore="12.2" y_score="10.7"
    // y_ions="9" b_score="0.0" b_ions="0"
    // pre="VLGR" post="VEFM" seq="TGSQGQCTQVR" missed_cleavages="10">

    // valeur généric du scan
    _current_peptide_sp = PeptideXtp(attributes.value("seq").simplified()).makePeptideXtpSp();
    _p_peptide_match = new PeptideMatch(_sp_msrun.get(), attributes.value("id").simplified().section(".",0,0).toUInt());

    _p_peptide_match->setRetentionTime(_retention_time);
    _p_peptide_match->setEvalue(attributes.value("expect").simplified().toDouble());
    //qDebug() << "XtandemSaxHandler::startElement_domain evalue "  << _p_peptide_match->getEvalue() << " scan " << _p_peptide_match->getScan();

    pappso::pappso_double xtandem_mhtheo = attributes.value("mh").simplified().toDouble();
    pappso::pappso_double xtandem_delta = attributes.value("delta").simplified().toDouble();

    pappso::pappso_double exp_mass = xtandem_mhtheo + xtandem_delta - pappso::MHPLUS;

    _p_peptide_match->setExperimentalMass(exp_mass);
    _p_peptide_match->setStart(attributes.value("start").simplified().toUInt()-1);
    _p_peptide_match->setCharge(_charge);

    _p_peptide_match->setIdentificationDataSource( _p_identification_data_source);
    _p_peptide_match->setChecked(true);
    _p_protein_match->addPeptideMatch(_p_peptide_match);



    // missing informations
    //peptide.set_hypercorr(Float.valueOf(attrs.getValue("hyperscore")));
    //peptide.set_pre(attrs.getValue("pre"));
    //peptide.set_post(attrs.getValue("post"));
    //qDebug() << "startElement_domain end" ;
    return is_ok;
}

bool XtandemSaxHandler::startElement_aa(QXmlAttributes attributes) {
//<aa type="M" at="624" modified="15.99491" />
    bool is_ok = true;
    //qDebug() << "startElement_aa ";
    pappso::AaModificationP modif = getAaModificationP(attributes.value("modified").simplified().toDouble());
    unsigned int position_in_prot = attributes.value("at").simplified().toUInt()-1;
    _current_peptide_sp.get()->addAaModification(modif, position_in_prot-_p_peptide_match->getStart());
    //qDebug() << "startElement_aa end" ;
    return is_ok;
}

bool XtandemSaxHandler::endElement_domain() {
   bool is_ok = true;
    _current_peptide_sp = _p_project->getPeptideStore().getInstance(_current_peptide_sp);

    _p_peptide_match->setPeptideXtpSp(_current_peptide_sp);
    return is_ok;
}
bool XtandemSaxHandler::endElement_note() {
//<note label="description">GRMZM2G083841_P01 P04711 Phosphoenolpyruvate carboxylase 1 (PEPCase 1)(PEPC 1)(EC //4.1.1.31) seq=translation; coord=9:61296279..61301686:1; parent_transcript=GRMZM2G083841_T01;
    ////parent_gene=GRMZM2G083841</note>
    bool is_ok = true;
    if (_is_protein_description) {
        _p_protein_match->getProteinXtpSp().get()->setDescription(_current_text.section(" ",1));
    }
    return is_ok;
}



bool XtandemSaxHandler::error(const QXmlParseException &exception) {
    _errorStr = QObject::tr("Parse error at line %1, column %2 :\n"
                            "%3").arg(exception.lineNumber()).arg(exception.columnNumber()).arg(
                    exception.message());

    return false;
}


bool XtandemSaxHandler::fatalError(const QXmlParseException &exception) {
    _errorStr = QObject::tr("Parse error at line %1, column %2 :\n"
                            "%3").arg(exception.lineNumber()).arg(exception.columnNumber()).arg(
                    exception.message());
    return false;
}

QString XtandemSaxHandler::errorString() const {
    return _errorStr;
}


bool XtandemSaxHandler::endDocument() {
    return true;
}

bool XtandemSaxHandler::startDocument() {
    return true;
}

bool XtandemSaxHandler::characters(const QString &str) {
    _current_text += str;
    return true;
}


pappso::AaModificationP XtandemSaxHandler::getAaModificationP(pappso::mz mass) const {
    pappso::PrecisionP precision = pappso::Precision::getDaltonInstance(0.01);

    pappso::AaModificationP oxidation = pappso::AaModification::getInstance("MOD:00719");
    if (pappso::MassRange(oxidation->getMass(),precision).contains(mass)) {
        return oxidation;
    }
    pappso::AaModificationP iodoacetamide = pappso::AaModification::getInstance("MOD:00397");
    if (pappso::MassRange(iodoacetamide->getMass(),precision).contains(mass)) {
        return iodoacetamide;
    }
    pappso::AaModificationP acetylated = pappso::AaModification::getInstance("MOD:00408");
    if (pappso::MassRange(acetylated->getMass(),precision).contains(mass)) {
        return acetylated;
    }
    pappso::AaModificationP phosphorylated = pappso::AaModification::getInstance("MOD:00696");
    if (pappso::MassRange(phosphorylated->getMass(),precision).contains(mass)) {
        return phosphorylated;
    }
    pappso::AaModificationP ammonia = pappso::AaModification::getInstance("MOD:01160");
    if (pappso::MassRange(ammonia->getMass(),precision).contains(mass)) {
        return ammonia;
    }
    pappso::AaModificationP dehydrated = pappso::AaModification::getInstance("MOD:00704");
    if (pappso::MassRange(dehydrated->getMass(),precision).contains(mass)) {
        return dehydrated;
    }
    pappso::AaModificationP dimethylated = pappso::AaModification::getInstance("MOD:00429");
    if (pappso::MassRange(dimethylated->getMass(),precision).contains(mass)) {
        return dimethylated;
    }

    pappso::AaModificationP dimethylated_medium = pappso::AaModification::getInstance("MOD:00552");
    if (pappso::MassRange(dimethylated_medium->getMass(),precision).contains(mass)) {
        return dimethylated_medium;
    }

    pappso::AaModificationP dimethylated_heavy = pappso::AaModification::getInstance("MOD:00638");
    if (pappso::MassRange(dimethylated_heavy->getMass(),precision).contains(mass)) {
        return dimethylated_heavy;
    }
    pappso::AaModificationP DimethylpyrroleAdduct = pappso::AaModification::getInstance("MOD:00628");
    if (pappso::MassRange(DimethylpyrroleAdduct->getMass(),precision).contains(mass)) {
        return DimethylpyrroleAdduct;
    }

    // modification not found, creating customized mod :
    return pappso::AaModification::getInstanceCustomizedMod(mass);

    throw pappso::ExceptionNotFound(QObject::tr("XpipSaxHandler::getAaModificationP => modification not found for mass %1").arg(mass));
}

